package seamCarving;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

public class ImageMatrix {

    private HashMap<PixelCoord, Integer> map;

    private ArrayList<PixelCoord> seam;

    private BufferedImage origImage;

    private int width;
    private int height;

    public ImageMatrix(BufferedImage image) {
        this.origImage = image;
        this.width = image.getWidth();
        this.height = image.getHeight();
        
        // Map from (x,y) pixel position to the pixel color
        this.map = new HashMap<PixelCoord, Integer>(this.width * this.height);
        for (int i = 0; i < this.width; i++) {
            for (int j = 0; j < this.height; j++) {
                int rgb = image.getRGB(i,j);
                this.setRGB(i, j, rgb);
            }
        }

        this.seam = null;
    }

    public BufferedImage getImage() {
        BufferedImage image = new BufferedImage(this.width, this.height, this.origImage.getType());

        for (int i = 0; i < this.width; i++) {
            for (int j = 0; j < this.height; j++) {
                int rgb = this.getRGB(i,j);
                image.setRGB(i, j, rgb);
            }
        }

        return image;
    }

    /**
     * Colors all pixels in the seam with the same color.
     */
    public void colorSeam() {
        if (this.seam == null) {
            return;
        }

        int rgb = new Color(255, 0, 0).getRGB();
        for (int idx = 0; idx < this.seam.size(); idx++) {
            this.setRGB(this.seam.get(idx).x, this.seam.get(idx).y, rgb);
        }
    }

    /**
     * Removes the currently stored "best" seam, if one exists, and shrinks
     * the image accordingly.
     */
    public void removeSeam() {
        if (this.seam == null) {
            return;
        }

        // Track whether each row has been seen before, to check
        // for repeats in the seam
        boolean[] isRepeatedRow = new boolean[this.height];
        for (int j = 0; j < this.height; j++) {
            isRepeatedRow[j] = false;
        }

        // For each pixel in the seam, remove that pixel, and slide
        // left all pixels to the right
        for (int idx = 0; idx < this.seam.size(); idx++) {
            int i = this.seam.get(idx).x;
            int j = this.seam.get(idx).y;

            // Check for invalid rows
            if (j < 0 || j >= this.height) {
                System.out.printf("Seam has nonexistent row %d\n", j);
            } else if (isRepeatedRow[j]) {
                System.out.printf("Seam has repeated row %d\n", j);
            }

            isRepeatedRow[j] = true;

            // Shift left all pixels to the right
            for (int ii = i; ii < this.width-1; ii++) {
                int rgb = this.getRGB(ii + 1, j);
                this.setRGB(ii, j, rgb);
            }

            // Remove this pixel
            this.map.remove(new PixelCoord(this.width-1, j));
        }

        // Check if any rows were missed
        for (int j = 0; j < this.height; j++) {
            if (!isRepeatedRow[j]) {
                System.out.printf("Seam missed row %d\n", j);
            }
        }
        
        // Decrease the width by 1
        this.width--;
    }

    /**
     * Finds the best seam using Dynamic Programming and the energy
     * measure of the pixels.
     */
    public void findBestSeam() {
        // TODO: Replace this with your code!
        this.chooseRandomColumnSeam();
//        this.chooseRandomSeam();
    }

    /**
     * Chooses a random column of the current image to be the seam.
     */
    private void chooseRandomColumnSeam() {
        this.seam = new ArrayList<PixelCoord>();

        Random r = new Random();
        int i = r.nextInt(this.width);
        for (int j = 0; j < this.height; j++) {
            this.seam.add(new PixelCoord(i,j));
        }
    }

    private void chooseRandomSeam() {
        this.seam = new ArrayList<PixelCoord>();
        
        Random r = new Random();
        int i = r.nextInt(this.width); // choose random starting column
        for (int j = 0; j < this.height; j++) {
            int di = r.nextInt(3) - 1;
            i = Math.min(this.width-1, Math.max(0, i + di)); // crop it to [0, width-1]
            this.seam.add(new PixelCoord(i,j));
        }
    }

    /**
     * Returns the "energy" of the pixel at (i,j), which gives the cost
     * of removing this pixel from the image.
     * @param i - the x-coordinate of the pixel
     * @param j - the y-coordinate of the pixel
     * @return The energy associated with pixel (i,j)
     */
    public int getEnergy(int i, int j) {
        if (i == 0 || j == 0 || i == this.width - 1 || j == this.height - 1) {
            // Return a large value on the edge
            return 10000;
        } else {
            // Similar to Sobel gradient magnitude
            return this.getDistance(this.getRGB(i-1,j), this.getRGB(i+1,j)) +
                   this.getDistance(this.getRGB(i,j-1), this.getRGB(i,j+1)) +
                   this.getDistance(this.getRGB(i-1,j-1), this.getRGB(i+1,j+1)) +
                   this.getDistance(this.getRGB(i+1,j-1), this.getRGB(i-1,j+1));
        }
    }

    /**
     * Calculates a distance metric between two pixel colors.
     * @return The "distance" between the two colors
     */
    private int getDistance(int rgb1, int rgb2) {
        Color c1 = new Color(rgb1);
        Color c2 = new Color(rgb2);

        int res = 0;
        res += Math.abs(c1.getRed() - c2.getRed());
        res += Math.abs(c1.getGreen() - c2.getGreen());
        res += Math.abs(c1.getBlue() - c2.getBlue());
        return res;
    }

    private void setRGB(int i, int j, int rgb) {
        this.map.put(new PixelCoord(i, j), rgb);
    }

    private int getRGB(int i, int j) {
        try {
            return this.map.get(new PixelCoord(i, j));
        } catch (NullPointerException e){
            System.out.printf("Failed for i=%d, j=%d\n", i, j);
            return 0;
        }
    }
}
